/* -*- mode: c++; c-basic-offset:4 -*-
    importcertificatecommand.cpp

    This file is part of Kleopatra, the KDE keymanager
    Copyright (c) 2007 Klarälvdalens Datakonsult AB

    Kleopatra is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kleopatra is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    In addition, as a special exception, the copyright holders give
    permission to link the code of this program with any edition of
    the Qt library by Trolltech AS, Norway (or with modified versions
    of Qt that use the same license as Qt), and distribute linked
    combinations including the two.  You must obey the GNU General
    Public License in all respects for all of the code used other than
    Qt.  If you modify this file, you may extend this exception to
    your version of the file, but you are not obligated to do so.  If
    you do not wish to do so, delete this exception statement from
    your version.
*/

#include <config-kleopatra.h>

#include "importcertificatecommand.h"
#include "command_p.h"

#include "utils/classify.h"

#include <kleo/cryptobackendfactory.h>
#include <kleo/importjob.h>

#include <gpgme++/global.h>
#include <gpgme++/importresult.h>

#include <KLocale>
#include <KMessageBox>
#include <KJob>

#include <QByteArray>
#include <QFile>
#include <QFileDialog>
#include <QPointer>
#include <QString>
#include <QWidget>

#include <memory>
#include <cassert>

using namespace GpgME;
using namespace Kleo;

class ImportCertificateCommand::Private : public Command::Private {
    friend class ::ImportCertificateCommand;
    ImportCertificateCommand * q_func() const { return static_cast<ImportCertificateCommand*>( q ); }
public:
    explicit Private( ImportCertificateCommand * qq, KeyListController * c );
    ~Private();

    bool ensureHaveFile();
    void startImport( const QByteArray& data );
    void importResult( const GpgME::ImportResult& );
    GpgME::Protocol checkProtocol( const QByteArray& data, const QString& filename ) const;
    void showError( const GpgME::Error& error );
    void showDetails( const ImportResult& result );

private:
    QPointer<ImportJob> importJob;
    QString filename;
};

ImportCertificateCommand::Private * ImportCertificateCommand::d_func() { return static_cast<Private*>(d.get()); }
const ImportCertificateCommand::Private * ImportCertificateCommand::d_func() const { return static_cast<const Private*>(d.get()); }

ImportCertificateCommand::Private::Private( ImportCertificateCommand * qq, KeyListController * c )
    : Command::Private( qq, c ), importJob( 0 )
{
    
}

ImportCertificateCommand::Private::~Private() {}



ImportCertificateCommand::ImportCertificateCommand( KeyListController * p )
    : Command( new Private( this, p ) )
{
    
}

ImportCertificateCommand::ImportCertificateCommand( QAbstractItemView * v, KeyListController * p )
    : Command( v, new Private( this, p ) )
{
    
}

ImportCertificateCommand::~ImportCertificateCommand() {}

#define d d_func()
#define q q_func()

QString ImportCertificateCommand::fileName() const
{
    return d->filename;
}

void ImportCertificateCommand::setFileName( const QString& fileName )
{
    d->filename = fileName;
}

void ImportCertificateCommand::doStart()
{
    if ( !d->ensureHaveFile() ) {
        emit canceled();
        d->finished();
        return;
    }
    //TODO: use KIO here
    QFile in( d->filename );
    if ( !in.open( QIODevice::ReadOnly ) ) {
        KMessageBox::error( d->view(), i18n( "Could not open file %1 for reading", d->filename ), i18n( "Certificate Import Failed" ) );
        d->finished();
        return;
    }
    d->startImport( in.readAll() ); 
    emit info( i18n( "Importing certificate..." ) );
}

bool ImportCertificateCommand::Private::ensureHaveFile()
{
   if ( filename.isNull() )
       filename = QFileDialog::getOpenFileName( view(), i18n( "Select Certificate File" ), QString(), i18n( "Certificates (*.asc *.pem *.der *.p7c *.p12)" )  );
   return !filename.isNull();
}

void ImportCertificateCommand::Private::showDetails( const ImportResult& res )
{
    // ### TODO: make a keylisting over Import::fingerprints(), then
    // ### highlight imported certificates in view(), or maybe in a new tab?

    const KLocalizedString normalLine = ki18n("<tr><td align=\"right\">%1</td><td>%2</td></tr>");
    const KLocalizedString boldLine = ki18n("<tr><td align=\"right\"><b>%1</b></td><td>%2</td></tr>");

    QStringList lines;
    lines.push_back( normalLine.subs( i18n("Total number processed:") )
                     .subs( res.numConsidered() ).toString() );
    lines.push_back( normalLine.subs( i18n("Imported:") )
                     .subs( res.numImported() ).toString() );
    if ( res.newSignatures() )
        lines.push_back( normalLine.subs( i18n("New signatures:") )
                         .subs( res.newSignatures() ).toString() );
    if ( res.newUserIDs() )
        lines.push_back( normalLine.subs( i18n("New user IDs:") )
                         .subs( res.newUserIDs() ).toString() );
    if ( res.numKeysWithoutUserID() )
        lines.push_back( normalLine.subs( i18n("Keys without user IDs:") )
                         .subs( res.numKeysWithoutUserID() ).toString() );
    if ( res.newSubkeys() )
        lines.push_back( normalLine.subs( i18n("New subkeys:") )
                         .subs( res.newSubkeys() ).toString() );
    if ( res.newRevocations() )
        lines.push_back( boldLine.subs( i18n("Newly revoked:") )
                         .subs( res.newRevocations() ).toString() );
    if ( res.notImported() )
        lines.push_back( boldLine.subs( i18n("Not imported:") )
                         .subs( res.notImported() ).toString() );
    if ( res.numUnchanged() )
        lines.push_back( normalLine.subs( i18n("Unchanged:") )
                         .subs( res.numUnchanged() ).toString() );
    if ( res.numSecretKeysConsidered() )
        lines.push_back( normalLine.subs( i18n("Secret keys processed:") )
                         .subs( res.numSecretKeysConsidered() ).toString() );
    if ( res.numSecretKeysImported() )
        lines.push_back( normalLine.subs( i18n("Secret keys imported:") )
                         .subs( res.numSecretKeysImported() ).toString() );
    if ( res.numSecretKeysConsidered() - res.numSecretKeysImported() - res.numSecretKeysUnchanged() > 0 )
        lines.push_back( boldLine.subs( i18n("Secret keys <em>not</em> imported:") )
                         .subs(  res.numSecretKeysConsidered()
                                 - res.numSecretKeysImported()
                                 - res.numSecretKeysUnchanged() ).toString() );
    if ( res.numSecretKeysUnchanged() )
        lines.push_back( normalLine.subs( i18n("Secret keys unchanged:") )
                         .subs( res.numSecretKeysUnchanged() ).toString() );
    
    KMessageBox::information( view(),
                              i18n( "<qt><p>Detailed results of importing %1:</p>"
                                    "<table>%2</table></qt>" ,
                                    filename, lines.join( QString() ) ),
                              i18n( "Certificate Import Result" ) );
}

void ImportCertificateCommand::Private::showError( const GpgME::Error& err )
{
    assert( err );
    assert( !err.isCanceled() );
    const QString msg = i18n( "<qt><p>An error occurred while trying "
                              "to import the certificate %1:</p>"
                              "<p><b>%2</b></p></qt>",
                              filename,
                              QString::fromLocal8Bit( err.asString() ) );
    KMessageBox::error( view(), msg, i18n( "Certificate Import Failed" ) );
}

void ImportCertificateCommand::Private::importResult( const GpgME::ImportResult& result )
{
    if ( result.error().code() ) {
        if ( result.error().isCanceled() )
            emit q->canceled();
        else
            showError( result.error() );
    } else {
        showDetails( result );
    }

    finished();
}

void ImportCertificateCommand::Private::startImport( const QByteArray& data )
{
    const GpgME::Protocol protocol = findProtocol( filename );
    if ( protocol == GpgME::UnknownProtocol ) { //TODO: might use exceptions here
        KMessageBox::error( view(), i18n( "Could not determine certificate type of %1.", filename ), i18n( "Certificate Import Failed" ) );
        finished();
        return;
    }

    std::auto_ptr<ImportJob> job( CryptoBackendFactory::instance()->protocol( protocol )->importJob() );
    assert( job.get() );
    connect( job.get(), SIGNAL(result(GpgME::ImportResult)),
             q, SLOT(importResult(GpgME::ImportResult)) );
    connect( job.get(), SIGNAL(progress(QString,int,int)), 
             q, SIGNAL(progress(QString,int,int)) );
    if ( const GpgME::Error err = job->start( data ) ) {
        showError( err );
        finished();
    } else if ( err.isCanceled() ) {
        emit q->canceled();
        finished();
    } else {
        importJob = job.release();
    }
}


void ImportCertificateCommand::doCancel()
{
    if ( d->importJob )
        d->importJob->slotCancel();
}

#undef d
#undef q

#include "moc_importcertificatecommand.cpp"

